package com.bocloud.fms.booter.controller;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.common.model.BocloudResult;
import com.bocloud.common.model.RequestContext;
import com.bocloud.common.utils.Common;
import com.bocloud.common.utils.ResultTools;
import com.bocloud.fms.engine.FlowEngine;
import com.bocloud.fms.entity.Attachment;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.MediaType;
import org.springframework.util.CollectionUtils;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;

import java.io.File;
import java.io.IOException;
import java.util.*;

/**
 * @author hufy
 */
@RestController
@Api(value = "流程审核节点", tags = "流程审核节点")
@RequestMapping({"/v1/nodes"})
public class FlowNodeController extends FlowBaseController {

    @Autowired
    FlowEngine flowEngine;
    @Value("${upload.storage.path:/home/cmp}")
    private String path;

    /**
     * 审批
     *
     * @param id
     * @param auditFlag
     * @param remark
     * @param applyParams
     * @param action      操作类型
     * @param files       附件
     * @param context
     * @return
     */
    @PatchMapping(value = "/{id}")
    @ApiOperation(tags = {"CMC", "CSC"}, value = "流程审核节点审批", response = BocloudResult.class, produces = MediaType.APPLICATION_JSON_VALUE)
    public BocloudResult audit(@PathVariable(Common.ID) Long id,
                               @RequestParam(value = "auditFlag") String auditFlag,
                               @RequestParam(value = "remark", required = false) String remark,
                               @RequestParam(value = "ApplyParams", required = false) String applyParams,
                               @RequestParam(value = Common.ACTION, required = false) String action,
                               @RequestParam(value = "attachments", required = false) MultipartFile[] files,
                               @Value(Common.REQ_CONTEXT) RequestContext context) {
        Map<String, Object> args = new HashMap<>();
        args.put("id", id);
        args.put("auditFlag", auditFlag);
        args.put("remark", remark);
        if (StringUtils.isNotBlank(applyParams)) {
            args.put("ApplyParams", applyParams);
        }
        if ("audit".equals(action)) {
            List<Attachment> attachments;
            try {
                attachments = this.upload(files);
            } catch (Exception e) {
                logger.error("上传附件异常", e);
                return new BocloudResult(false, "上传附件异常");
            }
            BocloudResult result = flowEngine.executeNode(id, context, args, attachments);
            if (result.isFailed()) {
                // 删除垃圾文件
                for (Attachment attachment : attachments) {
                    if (new File(path, attachment.getPath()).exists()) {
                        boolean flag = new File(path, attachment.getPath()).delete();
                        logger.info("临时文件删除状态：{}", flag);
                    }
                }
            }
            return result;
        } else {
            return ResultTools.formatErrResult();
        }
    }

    /**
     * 批量审批
     *
     * @param instanceIds 实例id集合
     * @param action      操作类型
     * @param files       附件
     * @param context
     * @return
     */
    @PatchMapping
    @ApiOperation(tags = {"CMC", "CSC"}, value = "流程审核节点批量审批", response = BocloudResult.class, produces = MediaType.APPLICATION_JSON_VALUE)
    public BocloudResult patch(@RequestParam(value = "instanceIds", required = false) String instanceIds,
                               @RequestParam(value = "auditFlag") String auditFlag,
                               @RequestParam(value = "remark", required = false) String remark,
                               @RequestParam(value = Common.ACTION, required = false) String action,
                               @RequestParam(value = "attachments", required = false) MultipartFile[] files,
                               @Value(Common.REQ_CONTEXT) RequestContext context) {
        List<Long> instanceIdList = JSONObject.parseArray(instanceIds, Long.class);
        if (CollectionUtils.isEmpty(instanceIdList)) {
            return ResultTools.formatErrResult();
        }
        if ("audit".equals(action)) {
            List<Attachment> attachments;
            try {
                attachments = this.upload(files);
            } catch (Exception e) {
                logger.error("上传附件异常", e);
                return new BocloudResult(false, "上传附件异常");
            }

            Map<String, Object> args = new HashMap<>();
            args.put("auditFlag", auditFlag);
            args.put("remark", remark);
            BocloudResult result = flowEngine.batchExecuteNode(instanceIdList, context, args, attachments);
            if (result.isFailed()) {
                // 删除垃圾文件
                for (Attachment attachment : attachments) {
                    if (new File(path, attachment.getPath()).exists()) {
                        boolean flag = new File(path, attachment.getPath()).delete();
                        logger.info("临时文件删除状态：{}", flag);
                    }
                }
            }
            return result;
        } else {
            return ResultTools.formatErrResult();
        }
    }

    /**
     * 获取文件扩展名
     *
     * @param filename
     * @return
     */
    private String getExtname(String filename) {
        int index = filename.lastIndexOf('.');
        if (index == -1) {
            return null;
        }
        return filename.substring(index + 1);
    }

    /**
     * 上传附件
     *
     * @param multipartFiles
     * @return
     * @throws Exception
     */
    private List<Attachment> upload(MultipartFile[] multipartFiles) throws Exception {
        List<Attachment> attachments = new ArrayList<>();
        if (multipartFiles == null) {
            return attachments;
        }
        List<File> files = new ArrayList<>();
        try {
            for (MultipartFile multipartFile : multipartFiles) {
                String extname = getExtname(multipartFile.getOriginalFilename());
                String filename = File.separator + UUID.randomUUID().toString().replaceAll("-", "") + (null == extname ? "" : "." + extname);
                File file = new File("attachment" + filename);
                File destFile = new File(path, file.getPath());
                Attachment attachment = new Attachment();
                attachment.setContentType(multipartFile.getContentType());
                attachment.setSize(multipartFile.getSize());
                attachment.setName(multipartFile.getOriginalFilename());
                attachment.setPath(file.getPath());
                if (!destFile.getParentFile().exists()) {
                    destFile.getParentFile().mkdirs();
                }
                multipartFile.transferTo(destFile);
                files.add(file);
                attachments.add(attachment);
            }
        } catch (IOException e) {
            // 删除垃圾文件
            for (File file : files) {
                if (new File(path, file.getPath()).exists()) {
                    boolean flag = new File(path, file.getPath()).delete();
                    logger.info("临时文件删除状态：{}", flag);
                }
            }
            throw new Exception(e);
        }
        return attachments;
    }
}
