package com.bocloud.fms.engine.impl;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.common.model.RequestContext;
import com.bocloud.fms.common.enums.FlowConstants;
import com.bocloud.fms.common.model.BasicDomainModel;
import com.bocloud.fms.engine.Completion;
import com.bocloud.fms.engine.freemarker.FreeMarkerUtil;
import com.bocloud.fms.engine.model.BaseNodeModel;
import com.bocloud.fms.engine.model.DefinitionModel;
import com.bocloud.fms.engine.model.FlowNodeModel;
import com.bocloud.fms.entity.FlowDefinition;
import com.bocloud.fms.entity.FlowInstance;
import com.bocloud.fms.entity.FlowNode;
import com.bocloud.fms.intf.FlowDefinitionService;
import com.bocloud.fms.intf.FlowQueryService;
import com.bocloud.fms.model.NotifyMessage;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.amqp.AmqpException;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.stereotype.Component;

/**
 * 默认的任务、实例完成时触发的动作
 *
 * @author hufy
 */
@Component("completion")
public class GeneralCompletion implements Completion {

    private static final Logger logger = LoggerFactory.getLogger(GeneralCompletion.class.getName());
    private final FlowQueryService flowQueryService;
    private final AmqpTemplate amqpTemplate;
    private final FlowDefinitionService definitionService;

    public GeneralCompletion(FlowQueryService flowQueryService, AmqpTemplate amqpTemplate,
                             FlowDefinitionService definitionService) {
        this.flowQueryService = flowQueryService;
        this.amqpTemplate = amqpTemplate;
        this.definitionService = definitionService;
    }

    @Override
    public void complete(FlowNode flowNode) {
        logger.info("The task[{}] has been user[{}] has completed", flowNode.getId(), flowNode.getOperatorId());
        FlowInstance flowInstance = flowQueryService.getFlowInstance(flowNode.getInstanceId());
        String variable = flowInstance.getVariable();
        JSONObject param = JSONObject.parseObject(variable);
        String title = param.getString(FlowConstants.APPLY_THEME);
        String name = flowQueryService.getDisplayName(
                new RequestContext(flowNode.getOperatorCatalog(), flowNode.getOperatorId(), null, null, null));
        String messageTitile = "【" + flowInstance.getInstanceNo() + "】《" + title + "》审核信息";

        String content = null;
        BasicDomainModel<FlowDefinition, DefinitionModel> definition =
                definitionService.getDefinition(flowInstance.getDefId());
        BaseNodeModel model = definition.getModel().getNode(flowNode.getNodeId());
        if (!(model instanceof FlowNodeModel)) {
            return;
        }
        // ApplyParams转换成json格式，并且将ApplyParams里的参数提到外面一个层级，为了方便生成消息模板
        if (param.getJSONObject(FlowConstants.APPLY_PARAMS) != null) {
            param.putAll(param.getJSONObject(FlowConstants.APPLY_PARAMS));
        }
        param.put("currentNode", flowNode.getName());
        param.put("operatorName", name);
        param.put("remark", flowNode.getVariableMap().get("remark"));
        if (FlowConstants.ARGS_AUDIT_FLAG_PASS.equals(flowNode.getVariableMap().get(FlowConstants.ARGS_AUDIT_FLAG))) {
            String template = ((FlowNodeModel) model).getMsgTemplateApproved();
            if (StringUtils.isBlank(template)) {
                return;
            }
            try {
                content = FreeMarkerUtil.getContent(template, param);
            } catch (Exception e) {
                logger.error("生成消息通知内容失败", e);
            }
        } else if (FlowConstants.ARGS_AUDIT_FLAG_NO_PASS
                .equals(flowNode.getVariableMap().get(FlowConstants.ARGS_AUDIT_FLAG))) {
            String template = ((FlowNodeModel) model).getMsgTemplateUnratified();
            if (StringUtils.isBlank(template)) {
                return;
            }
            try {
                content = FreeMarkerUtil.getContent(template, param);
            } catch (Exception e) {
                logger.error("生成消息通知内容失败", e);
            }
        } else {
            logger.warn("unkonwn auditFlag:" + flowNode.getVariableMap().get(FlowConstants.ARGS_AUDIT_FLAG));
            return;
        }

        String applyCategory = param.getString(FlowConstants.APPLY_CATEGORY);
        // F5申请内容信息
        if (FlowConstants.APPLY_CATEGORY_F5_APPLY.equals(applyCategory)) {
            JSONObject applyParams = param.getJSONObject(FlowConstants.APPLY_PARAMS);
            JSONArray members = applyParams.getJSONArray("members");
            content += "\n";
            content += "Pool Member:\n";
            for (int i = 0; i < members.size(); i++) {
                JSONObject member = members.getJSONObject(i);
                content += member.getString("poolMemberAddress") + ":" + member.getString("poolMemberPort");
                if (i != member.size() - 1) {
                    content += "\n";
                }
            }
        } else if (FlowConstants.APPLY_CATEGORY_F5_EXPANSION_APPLY.equals(applyCategory)) { // F5扩容申请内容信息
            JSONObject applyParams = param.getJSONObject(FlowConstants.APPLY_PARAMS);
            JSONArray members = applyParams.getJSONArray("members");
            content += "\n";
            content += "VirtualServers:" + applyParams.getJSONObject("virtualServer").getString("name") + "\n";
            content += "新增Pool Member:\n";
            for (int i = 0; i < members.size(); i++) {
                JSONObject member = members.getJSONObject(i);
                content += member.getString("poolMemberAddress") + ":" + member.getString("poolMemberPort");
                if (i != member.size() - 1) {
                    content += "\n";
                }
            }
        }
        NotifyMessage message = new NotifyMessage(
                new RequestContext(flowInstance.getCreatorCatalog(), flowInstance.getCreatorId(), 0L, null, null),
                messageTitile, content, "<pre>" + content + "</pre>", true, false);
        amqpTemplate.convertAndSend("event.notice.queue", message);
    }

    @Override
    public void complete(FlowInstance flowInstance) {
        logger.info("The flowInstance[{}] has completed", flowInstance.getId());
        String variable = flowInstance.getVariable();
        JSONObject paramObj = JSONObject.parseObject(variable);
        String category = paramObj.getString(FlowConstants.APPLY_CATEGORY);
        try {
            paramObj.put(FlowConstants.APPLY_STATUS, FlowConstants.APPLY_STATUS_APPROVED);
            // 服务目录
            if (FlowConstants.APPLY_CATEGORY_SERVICE_RELEASE.equals(category)
                    || FlowConstants.APPLY_CATEGORY_SERVICE_UNRELEASE.equals(category)
                    || FlowConstants.APPLY_CATEGORY_SERVICE_STOP.equals(category)
                    || FlowConstants.APPLY_CATEGORY_SERVICE_APPLY.equals(category)
                    || FlowConstants.APPLY_CATEGORY_SERVICE_POSTPONE.equals(category)) {
                amqpTemplate.convertAndSend(FlowConstants.SERVICE_APPLY_QUEUE, paramObj);
            } else if (FlowConstants.APPLY_CATEGORY_COMMON_TASK_START.equals(category)
                    || FlowConstants.APPLY_CATEGORY_SIMPLE_TASK_START.equals(category)
                    || FlowConstants.APPLY_CATEGOYR_CRON_TASK_START.equals(category)
                    || FlowConstants.APPLY_CATEGORY_SCRIPT_UPLOAD.equals(category)
                    || FlowConstants.APPLY_CATEGOYR_SOFTWARE_UPLOAD.equals(category)
                    || FlowConstants.APPLY_CATEGORY_RESOURCE_COMPUTER_APPLY.equals(category)
                    || FlowConstants.APPLY_CATEGORY_RESOURCE_IMAGE_APPLY.equals(category)
                    || FlowConstants.APPLY_CATEOGRY_RESOURCE_STORAGE_APPLY.equals(category)
                    || FlowConstants.APPLY_CATEGORY_RESOURCE_SNAPSHOT_APPLY.equals(category)
                    || FlowConstants.APPLY_CATEGORY_RESOURCE_ROUTE_APPLY.equals(category)
                    || FlowConstants.APPLY_CATEGORY_RESOURCE_FLOATING_IP_APPLY.equals(category)) {
                amqpTemplate.convertAndSend(FlowConstants.BOCLOUD_FLOW_QUEUE, paramObj);
            } else if (FlowConstants.APPLY_CATEGORY_RESOURCE_SHOPPING_APPLY.equals(category)) {
                JSONArray jsonArray = paramObj.getJSONArray(FlowConstants.APPLY_PARAMS);
                for (JSONObject jsonObject : jsonArray.toJavaList(JSONObject.class)) {
                    jsonObject.put(FlowConstants.APPLY_STATUS, FlowConstants.APPLY_STATUS_APPROVED);
                    amqpTemplate.convertAndSend(FlowConstants.BOCLOUD_FLOW_QUEUE, jsonObject);
                }
            } else {
                amqpTemplate.convertAndSend(FlowConstants.BOCLOUD_FLOW_QUEUE, paramObj);
            }
        } catch (AmqpException e) {
            logger.error("发送通知失败", e);
        }
    }

}
