package com.bocloud.fms.engine.model;

import com.bocloud.fms.common.enums.FlowConstants;
import com.bocloud.fms.common.exception.FlowException;
import com.bocloud.fms.common.utils.ClassUtil;
import com.bocloud.fms.common.utils.StringUtil;
import com.bocloud.fms.engine.Action;
import com.bocloud.fms.engine.Expression;
import com.bocloud.fms.engine.FlowInterceptor;
import com.bocloud.fms.engine.spring.SpelExpression;
import com.bocloud.fms.service.Execution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * 节点元素（存在输入输出的变迁）
 *
 * @author hufy
 */
public abstract class BaseNodeModel extends BaseModel implements Action {
    /**
     *
     */
    private static final long serialVersionUID = -2377317472320109317L;
    private static final Logger log = LoggerFactory.getLogger(BaseNodeModel.class);

    protected Expression expression = new SpelExpression();

    /**
     * 编码
     */
    private String code;
    /**
     * 输入变迁集合
     */
    private List<TransitionModel> inputs = new ArrayList<>();
    /**
     * 输出变迁集合
     */
    private List<TransitionModel> outputs = new ArrayList<>();
    /**
     * layout
     */
    private String layout;
    /**
     * 局部前置拦截器
     */
    private String preInterceptors;
    /**
     * 局部后置拦截器
     */
    private String postInterceptors;
    /**
     * 前置局部拦截器实例集合
     */
    private List<FlowInterceptor> preInterceptorList = new ArrayList<>();
    /**
     * 后置局部拦截器实例集合
     */
    private List<FlowInterceptor> postInterceptorList = new ArrayList<>();

    /**
     * 具体节点模型需要完成的执行逻辑
     *
     * @param execution 执行对象
     */
    protected abstract void exec(Execution execution);

    /**
     * 对执行逻辑增加前置、后置拦截处理
     *
     * @param execution 执行对象
     */
    @Override
    public void execute(Execution execution) {
        intercept(preInterceptorList, execution);
        exec(execution);
        intercept(postInterceptorList, execution);
    }

    /**
     * 运行变迁继续执行
     *
     * @param execution 执行对象
     */
    protected void runOutTransition(Execution execution) {
        for (TransitionModel tm : getOutputs()) {
            if (StringUtil.isNotEmpty(tm.getExpression())) {
                boolean result = expression.eval(Boolean.class, tm.getExpression(), execution.getArgs());
                if (!result) {
                    continue;
                }
            }
            tm.setEnabled(true);
            tm.execute(execution);
        }
    }

    /**
     * 拦截方法
     *
     * @param interceptorList 拦截器列表
     * @param execution       执行对象
     */
    private void intercept(List<FlowInterceptor> interceptorList, Execution execution) {
        try {
            for (FlowInterceptor interceptor : interceptorList) {
                interceptor.intercept(execution);
            }
        } catch (Exception e) {
            log.error("拦截器执行失败=" + e.getMessage());
            throw new FlowException(e);
        }
    }

    /**
     * 根据父节点模型、当前节点模型判断是否可退回。可退回条件： 1、满足中间无fork、join、subdefinition模型
     * 2、满足父节点模型如果为任务模型时，参与类型为any
     *
     * @param parent 父节点模型
     * @return 是否可以退回
     */
    public static boolean canRejected(BaseNodeModel current, BaseNodeModel parent) {
        boolean result = false;
        for (TransitionModel tm : current.getInputs()) {
            BaseNodeModel source = tm.getSource();
            if (source == parent) {
                return true;
            }
            if (source instanceof ForkModel || source instanceof JoinModel || source instanceof StartModel) {
                continue;
            }
            result = result || canRejected(source, parent);
        }
        return result;
    }

    public <T> List<T> getNextModels(Class<T> clazz) {
        List<T> models = new ArrayList<T>();
        for (TransitionModel tm : this.getOutputs()) {
            addNextModels(models, tm, clazz);
        }
        return models;
    }

    @SuppressWarnings("unchecked")
    protected <T> void addNextModels(List<T> models, TransitionModel tm, Class<T> clazz) {
        if (clazz.isInstance(tm.getTarget())) {
            models.add((T) tm.getTarget());
        } else {
            for (TransitionModel tm2 : tm.getTarget().getOutputs()) {
                addNextModels(models, tm2, clazz);
            }
        }
    }

    public String getCode() {
        return code;
    }

    public void setCode(String code) {
        this.code = code;
    }

    public List<TransitionModel> getInputs() {
        return inputs;
    }

    public void setInputs(List<TransitionModel> inputs) {
        this.inputs = inputs;
    }

    public List<TransitionModel> getOutputs() {
        return outputs;
    }

    public void setOutputs(List<TransitionModel> outputs) {
        this.outputs = outputs;
    }

    public String getLayout() {
        return layout;
    }

    public void setLayout(String layout) {
        this.layout = layout;
    }

    public String getPreInterceptors() {
        return preInterceptors;
    }

    public void setPreInterceptors(String preInterceptors) {
        this.preInterceptors = preInterceptors;
        if (StringUtil.isNotEmpty(preInterceptors)) {
            for (String interceptor : preInterceptors.split(FlowConstants.SPLIT_KEY_COMMA)) {
                FlowInterceptor instance = (FlowInterceptor) ClassUtil.newInstance(interceptor);
                if (instance != null) {
                    this.preInterceptorList.add(instance);
                }
            }
        }
    }

    public String getPostInterceptors() {
        return postInterceptors;
    }

    public void setPostInterceptors(String postInterceptors) {
        this.postInterceptors = postInterceptors;
        if (StringUtil.isNotEmpty(postInterceptors)) {
            for (String interceptor : postInterceptors.split(FlowConstants.SPLIT_KEY_COMMA)) {
                FlowInterceptor instance = (FlowInterceptor) ClassUtil.newInstance(interceptor);
                if (instance != null) {
                    this.postInterceptorList.add(instance);
                }
            }
        }
    }
}
