package com.bocloud.fms.engine.spring;

import com.bocloud.fms.engine.Context;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.DefaultListableBeanFactory;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.context.ApplicationContext;

import java.util.ArrayList;
import java.util.List;

/**
 * spring的服务查找实现
 *
 * @author yuqs
 * @since 1.5
 */
public class SpringContext implements Context {
    private static final Logger log = LoggerFactory.getLogger(SpringContext.class);
    /**
     * Spring context
     */
    private ApplicationContext applicationContext;

    private DefaultListableBeanFactory beanFactory;

    /**
     * 根据spring的上下文构造SpringContext
     *
     * @param ctx 上下文
     */
    public SpringContext(ApplicationContext ctx) {
        this.applicationContext = ctx;
        beanFactory = (DefaultListableBeanFactory) ctx.getAutowireCapableBeanFactory();
    }

    @Override
    @SuppressWarnings("unchecked")
    public <T> T find(Class<T> clazz) {
        String[] names = applicationContext.getBeanNamesForType(clazz);
        if (names.length > 1 && log.isWarnEnabled()) {
            log.warn("重复定义类型:" + clazz);
        }

        if (names.length >= 1) {
            return (T) applicationContext.getBean(names[0]);
        }
        try {
            T bean = applicationContext.getBean(clazz.getName(), clazz);
            if (bean != null) {
                return bean;
            }
        } catch (BeansException e) {
            log.info("没有找到bean" + clazz.getName() + "返回空");
        }
        return null;
    }

    @Override
    @SuppressWarnings("unchecked")
    public <T> List<T> findList(Class<T> clazz) {
        String[] names = applicationContext.getBeanNamesForType(clazz);
        List<T> beans = new ArrayList<T>();
        for (String name : names) {
            beans.add((T) applicationContext.getBean(name));
        }
        return beans;
    }

    @Override
    public <T> T findByName(String name, Class<T> clazz) {
        return applicationContext.getBean(name, clazz);
    }

    /**
     * spring不支持向applicationContext中直接添加对象
     */
    @Override
    public void put(String name, Object object) {
        log.warn("spring不支持向applicationContext中直接添加对象");
    }

    /**
     * 向spring添加bean的定义
     */
    @Override
    public void put(String name, Class<?> clazz) {
        BeanDefinition definition = new RootBeanDefinition(clazz, AutowireCapableBeanFactory.AUTOWIRE_BY_TYPE, true);
        beanFactory.registerBeanDefinition(name, definition);
        Object bean = beanFactory.getBean(clazz);
        log.info("添加进去了bean了" + bean.getClass().getName());
    }

    /**
     * 判断是否存在服务
     */
    @Override
    public boolean exist(String name) {
        return applicationContext.containsBean(name);
    }
}
