package com.bocloud.fms.service;

import com.bocloud.common.model.*;
import com.bocloud.common.utils.Common;
import com.bocloud.common.utils.ListTool;
import com.bocloud.common.utils.MapTools;
import com.bocloud.fms.common.enums.FlowConstants;
import com.bocloud.fms.common.exception.FlowException;
import com.bocloud.fms.common.model.BasicDomainModel;
import com.bocloud.fms.common.utils.AbstractAssertUtil;
import com.bocloud.fms.common.utils.StringUtil;
import com.bocloud.fms.engine.cache.Cache;
import com.bocloud.fms.engine.cache.CacheManager;
import com.bocloud.fms.engine.cache.CacheManagerAware;
import com.bocloud.fms.engine.model.AuditTaskModel;
import com.bocloud.fms.engine.model.DefinitionModel;
import com.bocloud.fms.engine.model.TransitionModel;
import com.bocloud.fms.engine.parser.ModelParser;
import com.bocloud.fms.entity.FlowDefinition;
import com.bocloud.fms.intf.FlowDefinitionService;
import com.bocloud.fms.model.DefinitionParams;
import com.bocloud.fms.repository.ServiceFlowConfigRepository;
import com.bocloud.fms.service.internal.SmsTenantService;
import com.bocloud.fms.service.internal.SmsUserService;
import com.bocloud.fms.internel.entity.Tenant;
import com.bocloud.fms.internel.entity.User;
import com.bocloud.service.lock.BocloudACLock;
import com.bocloud.service.lock.LockFactory;
import com.google.common.collect.Maps;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * 流程定义业务类
 *
 * @author hufy
 */
@Service
public class DefinitionServiceImpl extends AbstractAccessService implements FlowDefinitionService, CacheManagerAware {
    private static final Logger log = LoggerFactory.getLogger(DefinitionServiceImpl.class);
    private static final String DEFAULT_SEPARATOR = ".";

    @Autowired
    LockFactory lockFactory;
    @Autowired
    private ServiceFlowConfigRepository serviceFlowConfigRepository;
    @Autowired
    SmsUserService smsUserService;
    @Autowired
    SmsTenantService smsTenantService;

    /**
     * 流程定义对象cache名称
     */
    private static final String CACHE_DOMAIN_NAME = "bocloud.flow.definition.domain";
    /**
     * cache manager
     */
    private CacheManager cacheManager;
    /**
     * 实体cache(key=baselineId+","+version,value=domain对象) , 存放流程定义对象
     */
    private Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache;

    @Override
    public void check(FlowDefinition definition, Long id) {
        Assert.notNull(definition, "指定的流程定义[id/name=" + id + "]不存在");
        Assert.isTrue(definition.getState() != null && !definition.getState().equals(FlowConstants.STATE_FINISH),
                "指定的流程定义[id/name=" + id + ",version=" + definition.getVersion() + "]为非活动状态");
    }

    /**
     * 保存definition实体对象
     */
    @Override
    @Transactional
    public void saveDefinition(BasicDomainModel<FlowDefinition, DefinitionModel> definition) {
        this.getFlowDefinitionRepository().saveDefinition(definition.getEntity());
    }

    /**
     * 根据流程定义xml的输入流解析为字节数组，保存至数据库中，并且put到缓存中
     *
     * @param definitionParams 定义输入流
     * @param context          当前用户
     */
    @Override
    @Transactional
    public Long deploy(DefinitionParams definitionParams, RequestContext context) {
        AbstractAssertUtil.notNull(definitionParams);
        DefinitionModel model = ModelParser.parse(definitionParams.getGraph());
        // 开始对每个节点的验证
        checkRule(model);
        // 增加名称唯一性校验
        Assert.isTrue(!this.isNameExsit(definitionParams.getName(), context.getTenant()), "该名称已经存在");
        FlowDefinition entity = new FlowDefinition();
        BasicDomainModel<FlowDefinition, DefinitionModel> domain = new BasicDomainModel<>(entity, model);
        entity.setVersion(1);
        entity.setType(definitionParams.getType());
        entity.setState(FlowConstants.STATE_ACTIVE);
        entity.setGraph(definitionParams.getGraph());
        entity.setGmtCreate(new Date());
        entity.setGmtModify(new Date());
        entity.setCreatorId(context.getTarget());
        entity.setMenderId(context.getTarget());
        entity.setMenderCatalog(context.getCatalog());
        entity.setCreatorCatalog(context.getCatalog());
        entity.setName(definitionParams.getName());
        entity.setRemark(definitionParams.getRemark());
        entity.setOperationRemark("用户:" + this.getFlowUserName(context) + "创建了流程定义。");
        entity.setDeleted(false);
        entity.setTenantId(context.getTenant());
        saveDefinition(domain);
        cache(domain);
        return domain.getEntity().getId();

    }

    private void checkRule(DefinitionModel model) {
        List<AuditTaskModel> auditTaskModels = model.getModels(AuditTaskModel.class);
        for (AuditTaskModel auditTaskModel : auditTaskModels) {
            List<TransitionModel> outputs = auditTaskModel.getOutputs();
            // 标志位, 节点是不是只有审核未通过的线
            Boolean onlyHaveUnpass = true;
            for (TransitionModel output : outputs) {
                if (FlowConstants.ARGS_AUDIT_FLAG_PASS.equals(output.getAuditFlag())) {
                    onlyHaveUnpass = false;
                }
            }
            Assert.isTrue(!onlyHaveUnpass, "存在不符合规则的节点" + auditTaskModel.getName() + ",它只有审核不通过的连线");
        }
    }

    /**
     * 根据流程定义id、xml的输入流解析为字节数组，保存至数据库中，并且重新put到缓存中
     *
     * @param definitionParams 定义输入流
     * @param context          当前用户
     */
    @Override
    @Transactional
    public BocloudResult redeploy(Long id, DefinitionParams definitionParams, RequestContext context) {
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("id", id), Sign.EQ));
        params.add(new Param(MapTools.simpleMap("tenantId", context.getTenant()), Sign.EQ));
        List<FlowDefinition> definitions = this.getFlowDefinitionRepository()
                .getDefinitions(new Pager(1, Integer.MAX_VALUE, params, null), false);
        Assert.isTrue(!ListTool.isEmpty(definitions), "未能找到该流程");
        FlowDefinition definition = definitions.get(0);
        AbstractAssertUtil.notNull(definition);
        BasicDomainModel<FlowDefinition, DefinitionModel> domain = new BasicDomainModel<>(definition);
        try {
            boolean changeName = !definition.getName().equals(definitionParams.getName());
            if (changeName) {
                // 增加名称唯一性校验
                Boolean isNameExsit = this.isNameExsit(definitionParams.getName(), context.getTenant());
                if (isNameExsit) {
                    throw new FlowException("该名称已经存在");
                }
            }
            DefinitionModel model = ModelParser.parse(definitionParams.getGraph());
            Long baseLineId = domain.getEntity().getBaseLineId();
            domain.setModel(model);
            definition.setGraph(definitionParams.getGraph());
            definition.setId(null);
            definition.setGmtModify(new Date());
            definition.setName(definitionParams.getName());
            definition.setType(definitionParams.getType());
            definition.setRemark(definitionParams.getRemark());
            // 根据用户id查询用户名称
            definition.setOperationRemark("用户:" + this.getFlowUserName(context) + "编辑了模板。");
            definition.setMenderId(context.getTarget());
            definition.setMenderCatalog(context.getCatalog());
            Integer oldVersion = definition.getVersion();
            String path = FlowDefinitionService.class.getSimpleName() + "_" + baseLineId;
            // 获取版本号
            try (BocloudACLock lock = lockFactory.getACLock(path)) {
                Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
                definition.setVersion(getLatestVersion(definition.getBaseLineId()) + 1);
                // 修改了名称
                if (changeName) {
                    this.getFlowDefinitionRepository().updateName(definition.getBaseLineId(), definitionParams.getName());
                }
                // 新增定义
                this.getFlowDefinitionRepository().saveDefinition(definition);
                // 修改历史版本的
                Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache = ensureAvailableDomainCache();
                if (domainCache != null) {
                    domainCache.remove(baseLineId + DEFAULT_SEPARATOR + oldVersion);
                }
                cache(domain);
                return new BocloudResult(true, definition, "保存成功");
            } catch (Exception e) {
//                log.error(ExceptionUtils.getFullStackTrace(e));
                log.error(e.getMessage());
                return new BocloudResult(false, "保存失败");
            }
        } catch (Exception e) {
//            log.error(ExceptionUtils.getFullStackTrace(e));
            log.error(e.getMessage());
            throw new FlowException(e.getMessage(), e.getCause());
        }
    }

    /**
     * 根据definitionId卸载流程
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean terminate(Long id, RequestContext context, String remark) {
        FlowDefinition current = this.flowDefinitionRepository.getDefinition(id);
        boolean isBinding = this.serviceFlowConfigRepository.exists(current.getBaseLineId());
        if (isBinding) {
            throw new FlowException("该流程已绑定服务，无法终止");
        }
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("baselineId", current.getBaseLineId()), Sign.EQ));
        params.add(new Param(MapTools.simpleMap("tenantId", context.getTenant()), Sign.EQ));
        // 查询这个流程的所有版本
        List<FlowDefinition> definitions = this.getFlowDefinitionRepository()
                .getDefinitions(new Pager(1, Integer.MAX_VALUE, params, null), false);
        Assert.notNull(definitions, "该流程不存在");

        String path = FlowDefinitionService.class.getSimpleName() + "_" + current.getBaseLineId();
        try (BocloudACLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
            for (int i = 0; i < definitions.size(); i++) {
                FlowDefinition definition = definitions.get(i);
                BasicDomainModel<FlowDefinition, DefinitionModel> domain = new BasicDomainModel<>();
                domain.setEntity(definition);
                definition.setState(FlowConstants.STATE_TERMINATION);
                definition.setMenderId(context.getTarget());
                definition.setMenderCatalog(context.getCatalog());
                definition.setGmtModify(new Date());
                this.getFlowDefinitionRepository().update(domain.getEntity());
                cache(domain);
                if (i == definitions.size() - 1) {
                    FlowDefinition cloneDefinition = definition.clone();
                    cloneDefinition.setId(null);
                    cloneDefinition.setVersion(getLatestVersion(definition.getBaseLineId()) + 1);
                    cloneDefinition.setOperationRemark("用户:" + this.getFlowUserName(context) + "终止了模板，" + "备注为:" + remark + "。");
                    this.getFlowDefinitionRepository().saveDefinition(cloneDefinition);
                    BasicDomainModel<FlowDefinition, DefinitionModel> cloneDomain = new BasicDomainModel<>();
                    cloneDomain.setEntity(cloneDefinition);
                    cache(cloneDomain);
                }
            }
            return true;
        } catch (Exception e) {
            throw new FlowException(e);
        }
    }

    @Override
    public Integer getLatestVersion(Long id) {
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("baselineId", id), Sign.EQ));
        List<FlowDefinition> latestDefinitions = this.getFlowDefinitionRepository()
                .getDefinitions(new Pager(1, Integer.MAX_VALUE, params, null), true);
        if (!ListTool.isEmpty(latestDefinitions)) {
            return latestDefinitions.get(0).getVersion();
        }
        return 1;
    }

    /**
     * 获取最新版本的定义
     *
     * @param id
     * @return 最新版本的定义
     */
    @Override
    public BasicDomainModel<FlowDefinition, DefinitionModel> getLatestDefinition(Long id) {
        FlowDefinition latestDefinition = this.getFlowDefinitionRepository().getLatestDefinition(id);
        BasicDomainModel<FlowDefinition, DefinitionModel> domainModel = null;
        String key = latestDefinition.getBaseLineId() + DEFAULT_SEPARATOR + latestDefinition.getVersion();
        BasicDomainModel<FlowDefinition, DefinitionModel> modelFromCache = getModelFromCache(key, domainCache);
        if (modelFromCache != null) {
            return modelFromCache;
        }
        if (latestDefinition != null) {
            domainModel = new BasicDomainModel<>(latestDefinition);
            cache(domainModel);
        }
        return domainModel;
    }

    @Override
    public BasicDomainModel<FlowDefinition, DefinitionModel> getDefinition(Long id) {
        AbstractAssertUtil.notEmpty(id);
        BasicDomainModel<FlowDefinition, DefinitionModel> domainModel = null;
        Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache = ensureAvailableDomainCache();
        FlowDefinition definition = this.getFlowDefinitionRepository().getDefinition(id);
        String key = definition.getBaseLineId() + DEFAULT_SEPARATOR + definition.getVersion();
        BasicDomainModel<FlowDefinition, DefinitionModel> modelFromCache = getModelFromCache(key, domainCache);
        if (modelFromCache != null) {
            return modelFromCache;
        }
        if (definition != null) {
            domainModel = new BasicDomainModel<>(definition);
            cache(domainModel);
        }
        return domainModel;
    }

    private BasicDomainModel<FlowDefinition, DefinitionModel> getModelFromCache(String key,
                                                                                Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache) {

        BasicDomainModel<FlowDefinition, DefinitionModel> domainModel = null;
        if (domainCache != null) {
            if (StringUtil.isNotEmpty(key)) {
                domainModel = domainCache.get(key);
            }
        }
        return domainModel;

    }

    @Override
    public int count(RequestContext context, Pager pager, boolean latest) {
        return this.getFlowDefinitionRepository().count(pager, latest);
    }

    @Override
    public List<FlowDefinition> getDefinitions(RequestContext context, Pager pager, boolean latest) {
        List<FlowDefinition> definitions = this.getFlowDefinitionRepository().getDefinitions(pager, latest);
        if (ListTool.isEmpty(definitions)) {
            return Collections.EMPTY_LIST;
        }
        definitions.forEach(definition -> {
            definition.setCreatorName(getFlowUserName(new RequestContext(definition.getCreatorCatalog(), definition.getCreatorId(), definition.getTenantId(), null, null)));
            definition.setMenderName(getFlowUserName(new RequestContext(definition.getMenderCatalog(), definition.getMenderId(), definition.getTenantId(), null, null)));
        });
        return definitions;
    }

    @Override
    public List<FlowDefinition> getDefinitionHistory(RequestContext context, Long id) {
        FlowDefinition definition = this.getFlowDefinitionRepository().getDefinition(id);
        Map<String, String> sorter = Maps.newHashMap();
        sorter.put(Common.ID, Common.ONE);
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("baselineId", definition.getBaseLineId()), Sign.EQ));
        return this.getDefinitions(context, new Pager(1, Integer.MAX_VALUE, params, sorter), false);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public BocloudResult remove(Long id, RequestContext context) {
        FlowDefinition definition = this.flowDefinitionRepository.getDefinition(id);
        boolean isBinding = this.serviceFlowConfigRepository.exists(definition.getBaseLineId());
        if (isBinding) {
            return new BocloudResult(false, "该流程已绑定服务，无法删除");
        }
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("baselineId", definition.getBaseLineId()), Sign.EQ));
        List<FlowDefinition> definitions = this.flowDefinitionRepository
                .getDefinitions(new Pager(1, Integer.MAX_VALUE, params, null), false);
        for (FlowDefinition flowDefinition : definitions) {
            flowDefinition.setDeleted(true);
            flowDefinition.setMenderId(context.getTarget());
            flowDefinition.setMenderCatalog(context.getCatalog());
            this.flowDefinitionRepository.update(flowDefinition);
        }
        return new BocloudResult(true, "删除成功");
    }

    @Override
    public void checkPermision(FlowDefinition entity, Long tenantId) {
        if (!tenantId.equals(entity.getTenantId())) {
            throw new FlowException("租户id不匹配");
        }
    }

    @Override
    public Boolean isNameExsit(String name, Long tenantId) {
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("name", name), Sign.EQ));
        params.add(new Param(MapTools.simpleMap("tenantId", tenantId), Sign.EQ));
        List<FlowDefinition> definitions = this.getFlowDefinitionRepository()
                .getDefinitions(new Pager(1, Integer.MAX_VALUE, params, null), true);
        return definitions != null && definitions.size() > 0;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean active(Long id, RequestContext context, String remark) {
        List<Param> params = new ArrayList<>();
        params.add(new Param(MapTools.simpleMap("id", id), Sign.EQ));
        params.add(new Param(MapTools.simpleMap("tenantId", context.getTenant()), Sign.EQ));
        List<FlowDefinition> definitions = this.getFlowDefinitionRepository()
                .getDefinitions(new Pager(1, Integer.MAX_VALUE, params, null), false);
        if (!ListTool.isEmpty(definitions)) {
            FlowDefinition definition = definitions.get(0);
            AbstractAssertUtil.notNull(definition);
            Assert.isTrue(FlowConstants.STATE_TERMINATION.equals(definition.getState()), "该流程不是已终止状态");
            Assert.isTrue(!definition.getDeleted(), "该流程已经删除");
            BasicDomainModel<FlowDefinition, DefinitionModel> domain = new BasicDomainModel<>(definition);
            try {
                Long baseLineId = domain.getEntity().getBaseLineId();
                definition.setId(null);
                definition.setGmtModify(new Date());
                definition.setOperationRemark("用户:" + this.getFlowUserName(context) + "重新激活了流程定义:" + remark + "。");
                definition.setState(FlowConstants.STATE_ACTIVE);
                definition.setMenderId(context.getTarget());
                definition.setMenderCatalog(context.getCatalog());
                Integer oldVersion = definition.getVersion();
                String path = FlowDefinitionService.class.getSimpleName() + "_" + baseLineId;
                try (BocloudACLock lock = lockFactory.getACLock(path)) {
                    Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
                    definition.setVersion(getLatestVersion(definition.getBaseLineId()) + 1);
                    // 新增定义
                    this.getFlowDefinitionRepository().saveDefinition(definition);
                    Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache = ensureAvailableDomainCache();
                    if (domainCache != null) {
                        domainCache.remove(baseLineId + DEFAULT_SEPARATOR + oldVersion);
                    }
                    cache(domain);
                    return true;
                } catch (Exception e) {
//                    log.error(ExceptionUtils.getFullStackTrace(e));
                    log.error(e.getMessage());
                    return false;
                }
            } catch (Exception e) {
                log.error("重新激活出错 " + e.getMessage());
                throw new FlowException(e.getMessage(), e.getCause());
            }
        }

        throw new FlowException("出现查询错误");
    }

    /**
     * 缓存实体
     *
     * @param domain 流程定义对象
     */
    private void cache(BasicDomainModel<FlowDefinition, DefinitionModel> domain) {
        Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache = ensureAvailableDomainCache();
        // 如果model 为空从数据库里读取字节码，解析model
        if (domain.getModel() == null && domain.getEntity().getGraph() != null) {
            domain.setModel(ModelParser.parse(domain.getEntity().getGraph()));
        }
        String key = domain.getEntity().getBaseLineId() + DEFAULT_SEPARATOR + domain.getEntity().getVersion();
        if (domainCache != null) {
            if (log.isDebugEnabled()) {
                log.debug("cache definition id is[{}],name is[{}]", domain.getEntity().getId(), key);
            }
            domainCache.put(key, domain);
        } else {
            if (log.isDebugEnabled()) {
                log.debug("no cache implementation class");
            }
        }
    }

    @Override
    public void setCacheManager(CacheManager cacheManager) {
        this.cacheManager = cacheManager;
    }

    private Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> ensureAvailableDomainCache() {
        Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> entityCache = ensureDomainCache();
        if (entityCache == null && this.cacheManager != null) {
            entityCache = this.cacheManager.getCache(CACHE_DOMAIN_NAME);
        }
        return entityCache;
    }

    public Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> ensureDomainCache() {
        return domainCache;
    }

    public void setDomainCache(Cache<String, BasicDomainModel<FlowDefinition, DefinitionModel>> domainCache) {
        this.domainCache = domainCache;
    }

    /**
     * 获取用户名称
     *
     * @param context
     * @return
     */
    private String getFlowUserName(RequestContext context) {
        if (RequestContext.Catalog.Tenant == context.getCatalog()) {
            Tenant tenant = smsTenantService.detail(context.getTarget());
            return tenant != null ? tenant.getName() : "";
        } else {
            BocloudResult<User> result = smsUserService.detail(context.getTarget());
            return result.isSuccess() ? result.getData().getName() : "";
        }
    }

}
