package com.bocloud.fms.service;

import com.bocloud.common.model.RequestContext;
import com.bocloud.fms.common.exception.FlowException;
import com.bocloud.fms.common.model.BasicDomainModel;
import com.bocloud.fms.engine.FlowEngine;
import com.bocloud.fms.engine.model.DefinitionModel;
import com.bocloud.fms.engine.model.FlowNodeModel;
import com.bocloud.fms.entity.FlowDefinition;
import com.bocloud.fms.entity.FlowInstance;
import com.bocloud.fms.entity.FlowNode;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * 流程执行过程中所传递的执行对象，充当载体的角色，其中包含流程定义、流程模型、流程实例对象、执行参数、返回的任务列表
 *
 * @author hufy
 */
public class Execution implements Serializable {

    /**
     *
     */
    private static final long serialVersionUID = 3730741790729624400L;
    /**
     * FlowEngine holder
     */
    private FlowEngine engine;
    /**
     * 流程定义对象
     */
    private BasicDomainModel<FlowDefinition, DefinitionModel> definition;
    /**
     * 流程实例对象
     */
    private FlowInstance flowInstance;
    /**
     * 父流程实例
     */
    private FlowInstance parentFlowInstance;
    /**
     * 父流程实例节点名称
     */
    private String parentNodeName;
    /**
     * 子流程实例节点名称
     */
    private Long childFlowInstanceId;
    /**
     * 执行参数
     */
    private Map<String, Object> args;
    /**
     * 当前登录用户
     */
    private RequestContext context;
    /**
     * 任务
     */
    private BasicDomainModel<FlowNode, FlowNodeModel> task;
    /**
     * 返回的待处理任务列表
     */
    private List<BasicDomainModel<FlowNode, FlowNodeModel>> tasks = new ArrayList<>();
    /**
     * 是否已合并 针对join节点的处理
     */
    private boolean isMerged = false;

    /**
     * 已经发消息给用户的节点
     */
    private List<BasicDomainModel<FlowNode, FlowNodeModel>> notifiedNodes = new ArrayList<>();

    /**
     * 用于产生子流程执行对象使用
     *
     * @param execution
     * @param definition
     * @param parentNodeName
     */
    Execution(Execution execution, BasicDomainModel<FlowDefinition, DefinitionModel> definition,
              String parentNodeName) {
        if (execution == null || definition == null || parentNodeName == null) {
            throw new FlowException("构造Execution对象失败，请检查execution、definition、parentNodeName是否为空");
        }
        this.engine = execution.getEngine();
        this.definition = definition;
        this.args = execution.getArgs();
        this.parentFlowInstance = execution.getFlowInstance();
        this.parentNodeName = parentNodeName;
        this.context = execution.getContext();
    }

    /**
     * 构造函数，接收流程定义、流程实例对象、执行参数
     *
     * @param definition
     * @param flowInstance
     * @param args
     */
    public Execution(FlowEngine engine, BasicDomainModel<FlowDefinition, DefinitionModel> definition,
                     FlowInstance flowInstance, Map<String, Object> args) {
        if (definition == null || flowInstance == null) {
            throw new FlowException("构造Execution对象失败，请检查definition、FlowInstance是否为空");
        }
        this.engine = engine;
        this.definition = definition;
        this.flowInstance = flowInstance;
        this.args = args;
    }

    /**
     * 根据当前执行对象execution、子流程定义definition、当前节点名称产生子流程的执行对象
     *
     * @param execution
     * @param definition
     * @param parentNodeName
     * @return
     */
    public Execution createSubExecution(Execution execution,
                                        BasicDomainModel<FlowDefinition, DefinitionModel> definition, String parentNodeName) {
        return new Execution(execution, definition, parentNodeName);
    }

    /**
     * 获取流程定义对象
     *
     * @return
     */
    public BasicDomainModel<FlowDefinition, DefinitionModel> getDefinition() {
        return definition;
    }

    /**
     * 获取流程模型对象
     *
     * @return
     */
    public DefinitionModel getModel() {
        return definition.getModel();
    }

    /**
     * 获取流程实例对象
     *
     * @return
     */
    public FlowInstance getFlowInstance() {
        return flowInstance;
    }

    /**
     * 获取执行参数
     *
     * @return
     */
    public Map<String, Object> getArgs() {
        return args;
    }

    /**
     * 返回任务结果集
     *
     * @return
     */
    public List<BasicDomainModel<FlowNode, FlowNodeModel>> getTasks() {
        return tasks;
    }

    /**
     * 添加任务集合
     *
     * @param tasks
     */
    public void addTasks(List<BasicDomainModel<FlowNode, FlowNodeModel>> tasks) {
        this.tasks.addAll(tasks);
    }

    public RequestContext getContext() {
        return context;
    }

    public void setContext(RequestContext context) {
        this.context = context;
    }

    /**
     * 返回任务
     *
     * @return
     */
    public BasicDomainModel<FlowNode, FlowNodeModel> getTask() {
        return task;
    }

    /**
     * 设置任务
     *
     * @param task
     */
    public void setTask(BasicDomainModel<FlowNode, FlowNodeModel> task) {
        this.task = task;
    }

    /**
     * 判断是否已经成功合并
     *
     * @return
     */
    public boolean isMerged() {
        return isMerged;
    }

    /**
     * 设置是否为已合并
     *
     * @param isMerged
     */
    public void setMerged(boolean isMerged) {
        this.isMerged = isMerged;
    }

    /**
     * 获取引擎
     *
     * @return
     */
    public FlowEngine getEngine() {
        return engine;
    }

    public FlowInstance getParentFlowInstance() {
        return parentFlowInstance;
    }

    public String getParentNodeName() {
        return parentNodeName;
    }

    public Long getChildFlowInstanceId() {
        return childFlowInstanceId;
    }

    public void setChildFlowInstanceId(Long childFlowInstanceId) {
        this.childFlowInstanceId = childFlowInstanceId;
    }

    public List<BasicDomainModel<FlowNode, FlowNodeModel>> getNotifiedNodes() {
        return notifiedNodes;
    }

    public void setNotifiedNodes(List<BasicDomainModel<FlowNode, FlowNodeModel>> notifiedNodes) {
        this.notifiedNodes = notifiedNodes;
    }
}
