package com.bocloud.fms.service;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.common.model.RequestContext;
import com.bocloud.fms.common.enums.FlowConstants;
import com.bocloud.fms.common.exception.FlowException;
import com.bocloud.fms.common.model.BasicDomainModel;
import com.bocloud.fms.common.utils.AbstractAssertUtil;
import com.bocloud.fms.common.utils.StringUtil;
import com.bocloud.fms.engine.AssignmentHandler;
import com.bocloud.fms.engine.Completion;
import com.bocloud.fms.engine.model.BaseNodeModel;
import com.bocloud.fms.engine.model.DefinitionModel;
import com.bocloud.fms.engine.model.FlowNodeModel;
import com.bocloud.fms.entity.FlowDefinition;
import com.bocloud.fms.entity.FlowInstance;
import com.bocloud.fms.entity.FlowNode;
import com.bocloud.fms.entity.FlowNodeActor;
import com.bocloud.fms.intf.FlowNodeService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.*;
import java.util.stream.Collectors;

/**
 * @author hufy 任务执行业务类
 */
@Service
public class NodeServiceImpl extends AbstractAccessService implements FlowNodeService {
    private static final Long START = -1L;

    @Autowired
    private AssignmentHandler genericAssignment;

    /**
     * 完成指定任务
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public BasicDomainModel<FlowNode, FlowNodeModel> complete(Long taskId, RequestContext context) {
        return complete(taskId, context, null);
    }

    /**
     * 完成指定任务 该方法仅仅结束活动任务，并不能驱动流程继续执行
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public BasicDomainModel<FlowNode, FlowNodeModel> complete(Long nodeId, RequestContext context, Map<String, Object> args) {
        FlowNode node = this.getFlowNodeRepository().getNode(nodeId);
        BasicDomainModel<FlowNode, FlowNodeModel> domain = new BasicDomainModel<>(node);
        AbstractAssertUtil.notNull(node, "指定的任务[id=" + nodeId + "]不存在");
        domain.getEntity().setVariable(JSONObject.toJSONString(args));
        if (!isAllowed(domain.getEntity(), context)) {
            throw new FlowException("当前参与者[" + context.getTarget() + "]不允许执行任务[taskId=" + nodeId + "]");
        }
        node.setGmtFinish(new Date());
        node.setNodeState(FlowConstants.STATE_FINISH);
        node.setOperatorId(context.getTarget());
        node.setOperatorCatalog(context.getCatalog() == RequestContext.Catalog.Tenant ? RequestContext.Catalog.Tenant : RequestContext.Catalog.User);
        if (node.getActors() == null) {
            List<FlowNodeActor> actors = this.flowNodeActorRepository.getTaskActorsByTaskId(node.getId());
            List<RequestContext> contexts = actors.stream().map(actor -> new RequestContext(actor.getActorCatalog(), actor.getId(), null, null, null))
                    .collect(Collectors.toList());
            node.setActors(contexts);
        }
        this.getFlowNodeRepository().update(node);
        if (!FlowConstants.AUTO.equals(context.getTarget())) {
            Completion completion = getCompletion();
            if (completion != null) {
                completion.complete(domain.getEntity());
            }
        }
        return domain;
    }

    /**
     * 更新任务对象的finish_Time、operator、expire_Time、version、variable
     *
     * @param task 任务对象
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public void updateTask(FlowNode task) {
        this.getFlowNodeRepository().update(task);
    }

    /**
     * 唤醒指定的历史任务
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowNode resume(Long taskId, Long operator) {
        FlowNode histTask = this.getFlowNodeRepository().getNode(taskId);
        AbstractAssertUtil.notNull(histTask, "指定的历史任务[id=" + taskId + "]不存在");
        boolean isAllowed = true;
        if (histTask.getOperatorId() != null) {
            isAllowed = histTask.getOperatorId().equals(operator);
        }
        if (isAllowed) {
            BasicDomainModel<FlowNode, FlowNodeModel> task = new BasicDomainModel<>(histTask);
            task = undoNode(task);
            return task.getEntity();
        } else {
            throw new FlowException("当前参与者[" + operator + "]不允许唤醒历史任务[taskId=" + taskId + "]");
        }
    }

    /**
     * 撤回指定的任务
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowNode withdrawTask(Long taskId, Long operator) {
        FlowNode hist = this.getFlowNodeRepository().getNode(taskId);
        AbstractAssertUtil.notNull(hist, "指定的历史任务[id=" + taskId + "]不存在");
        List<FlowNode> tasks;
        tasks = this.getFlowNodeRepository().getNextActiveNodes(hist.getInstanceId(), hist.getNodeId(),
                hist.getParentNodeId());
        if (tasks == null || tasks.isEmpty()) {
            throw new FlowException("后续活动任务已完成或不存在，无法撤回.");
        }
        for (FlowNode task : tasks) {
            task.setDeleted(true);
            this.getFlowNodeRepository().update(task);
        }
        BasicDomainModel<FlowNode, FlowNodeModel> task = new BasicDomainModel<>(hist);
        task = undoNode(task);
        return task.getEntity();
    }

    /**
     * 驳回任务
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public BasicDomainModel<FlowNode, FlowNodeModel> rejectNode(DefinitionModel model, FlowNode currentTask) {
        Long parentTaskId = currentTask.getParentNodeId();
        if (parentTaskId == null || parentTaskId.equals(START)) {
            throw new FlowException("上一步任务ID为空，无法驳回至上一步处理");
        }
        BaseNodeModel current = model.getNode(currentTask.getNodeId());

        FlowNode histTask = this.getFlowNodeRepository().getNode(parentTaskId);
        BasicDomainModel<FlowNode, FlowNodeModel> history = new BasicDomainModel<>(histTask);
        BaseNodeModel parent = model.getNode(history.getEntity().getNodeId());
        if (!BaseNodeModel.canRejected(current, parent)) {
            throw new FlowException("无法驳回至上一步处理，请确认上一步骤并非fork、join任务");
        }

        BasicDomainModel<FlowNode, FlowNodeModel> task = new BasicDomainModel<>(histTask);
        task = undoNode(task);
        return task;
    }

    /**
     * 对指定的任务分配参与者
     *
     * @param task   任务
     * @param actors 参与者集合
     */
    @Transactional(rollbackFor = Exception.class)
    public void assignTask(FlowNode task, List<RequestContext> actors) {
        if (actors == null || actors.size() == 0) {
            return;
        }
        for (RequestContext actor : actors) {
            // 修复当actorId为null的bug
            if (actor == null) {
                continue;
            }
            FlowNodeActor taskActor = new FlowNodeActor();
            taskActor.setNodeId(task.getId());
            taskActor.setActorId(actor.getTarget());
            taskActor.setActorCatalog(actor.getCatalog());
            taskActor.setInstanceId(task.getInstanceId());
            taskActor.setDeleted(false);
            this.getFlowNodeActorRepository().save(taskActor);
        }
    }

    /**
     * 获取任务模型
     *
     * @param taskId 任务id
     * @return FlowNodeModel
     */
    @Override
    public FlowNodeModel getTaskModel(Long taskId) {
        FlowNode task = this.getFlowNodeRepository().getNode(taskId);
        AbstractAssertUtil.notNull(task);
        FlowInstance flowInstance = this.getFlowInstanceRepository().query(task.getInstanceId());
        AbstractAssertUtil.notNull(flowInstance);
        BasicDomainModel<FlowDefinition, DefinitionModel> definition = AbstractServiceContext.getEngine().definition()
                .getDefinition(flowInstance.getDefId());
        DefinitionModel model = definition.getModel();
        BaseNodeModel nodeModel = model.getNode(task.getNodeId());
        AbstractAssertUtil.notNull(nodeModel, "任务id无法找到节点模型.");
        if (nodeModel instanceof FlowNodeModel) {
            return (FlowNodeModel) nodeModel;
        } else {
            throw new IllegalArgumentException("任务id找到的节点模型不匹配");
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public BasicDomainModel<FlowNode, FlowNodeModel> undoNode(BasicDomainModel<FlowNode, FlowNodeModel> task) { // TODO
        return null;
    }

    /**
     * 由DBAccess实现类创建task，并根据model类型决定是否分配参与者
     *
     * @param flowNodeModel 模型
     * @param execution     执行对象
     * @return List<Task> 任务列表
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public List<BasicDomainModel<FlowNode, FlowNodeModel>> createTask(FlowNodeModel flowNodeModel,
                                                                      Execution execution) {
        List<BasicDomainModel<FlowNode, FlowNodeModel>> tasks = new ArrayList<>();

        Map<String, Object> args = execution.getArgs();
        if (args == null) {
            args = new HashMap<>(20);
        }
        String form = (String) args.get(flowNodeModel.getForm());
        String actionUrl = StringUtil.isEmpty(form) ? flowNodeModel.getForm() : form;

        Set<RequestContext> assignees = genericAssignment.getAssignees(flowNodeModel, execution.getFlowInstance().getProjectId(), execution.getContext());
        //没有审批人，自动通过，这里不再校验
//        Assert.isTrue(assignees != null && assignees.size() > 0, "流程审批人定义错误，未找到下一环节审批人，请联系管理员");
        List<RequestContext> assigneeContexts = new ArrayList<>(assignees);
        args.put(FlowNode.KEY_ACTOR, JSONObject.toJSONString(assigneeContexts));
        // 存入数据库
        BasicDomainModel<FlowNode, FlowNodeModel> domain = createTaskBase(flowNodeModel, execution);
        domain.getEntity().setActionUrl(actionUrl);
        domain.getEntity().setVariable(JSONObject.toJSONString(args));

        FlowNode flowTask = saveTask(domain.getEntity(), assigneeContexts);
        domain = new BasicDomainModel<>(flowTask);
        domain.setModel(this.getTaskModel(flowTask.getId()));
        tasks.add(domain);
        return tasks;
    }

    /**
     * 根据模型、执行对象、任务类型构建基本的task对象
     *
     * @param model     模型
     * @param execution 执行对象
     * @return Task任务对象
     */
    @Transactional(rollbackFor = Exception.class)
    // sonar : 非public方法，事务注解不会生效，故protected => public
    public BasicDomainModel<FlowNode, FlowNodeModel> createTaskBase(FlowNodeModel model, Execution execution) {
        BasicDomainModel<FlowNode, FlowNodeModel> task = new BasicDomainModel<>(new FlowNode(), model);
        task.getEntity().setInstanceId(execution.getFlowInstance().getId());
        task.getEntity().setNodeId(model.getId());
        task.getEntity().setName(model.getName());
        task.getEntity().setGmtCreate(new Date());
        task.getEntity().setVersion(execution.getFlowInstance().getVersion());
        task.getEntity().setNodeState(FlowConstants.STATE_ACTIVE);
        task.getEntity().setParentNodeId(execution.getTask() == null ? START : execution.getTask().getEntity().getId());
        task.setModel(model);
        return task;
    }

    /**
     * 由DBAccess实现类持久化task对象
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public FlowNode saveTask(FlowNode task, List<RequestContext> actors) {
        this.getFlowNodeRepository().save(task);
        assignTask(task, actors);
        task.setActors(actors);
        return task;
    }

    @Override
    public boolean isAllowed(FlowNode task, RequestContext operator) {
        if (operator != null) {
            if (FlowConstants.ADMIN.equals(operator.getTarget()) || FlowConstants.AUTO.equals(operator.getTarget())) {
                return true;
            }
            if (task.getOperatorId() != null) {
                return operator.getTarget().equals(task.getOperatorId()) && ((operator.getCatalog() == task.getOperatorCatalog()) || (operator.getCatalog() != RequestContext.Catalog.Tenant && task.getOperatorCatalog() != RequestContext.Catalog.Tenant));
            }
        }
        List<FlowNodeActor> actors = this.getFlowNodeActorRepository().getTaskActorsByTaskId(task.getId());
        if (actors == null || actors.isEmpty()) {
            return true;
        }
        if (operator == null) {
            return false;
        }
        boolean isAllowed = false;
        for (FlowNodeActor actor : actors) {
            if (actor.getActorId().equals(operator.getTarget())
                    && ((actor.getActorCatalog().equals(operator.getCatalog())) || (actor.getActorCatalog() != RequestContext.Catalog.Tenant && operator.getCatalog() != RequestContext.Catalog.Tenant))) {
                isAllowed = true;
                break;
            }
        }
        return isAllowed;
    }

    @Override
    public FlowNode getFlowNode(Long taskId) {
        return getFlowNodeRepository().getNode(taskId);
    }

}
