package com.bocloud.fms.service.internal;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.common.enums.BoCloudMethod;
import com.bocloud.common.enums.BoCloudService;
import com.bocloud.common.model.BocloudResult;
import com.bocloud.common.model.BocloudToken;
import com.bocloud.common.model.RequestContext.Catalog;
import com.bocloud.common.utils.Common;
import com.bocloud.common.utils.MapTools;
import com.bocloud.fms.internel.entity.Project;
import com.bocloud.fms.internel.entity.Tenant;
import com.bocloud.service.core.CurrentService;
import com.bocloud.service.core.Service;
import com.bocloud.service.http.core.ServiceFactory;
import com.bocloud.service.http.model.RemoteService;
import com.google.common.collect.Lists;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author hufy
 */
@Component
public class SmsProjectService implements InitializingBean {

    private final ServiceFactory serviceFactory;
    private final CurrentService currentService;

    @Autowired
    public SmsProjectService(ServiceFactory serviceFactory, CurrentService currentService) {
        this.serviceFactory = serviceFactory;
        this.currentService = currentService;
    }

    private final Map<String, Object> headers = new HashMap<>(4);

    private static final Service SERVICE = Service.create(BoCloudService.SMS);

    /**
     * 获取项目详情
     *
     * @param id 项目ID
     * @return 项目详情
     */
    public BocloudResult<Project> detail(Long id) {
        RemoteService remoteService = serviceFactory.build(SERVICE, "/v1/projects/" + id, BoCloudMethod.GET, headers,
                null);
        BocloudResult result = remoteService.invoke();
        Assert.isTrue(result.isSuccess(), result.getMessage());
        String data = JSONObject.toJSONString(result.getData());
        return new BocloudResult<>(true, JSONObject.parseObject(data, Project.class), "success");
    }

    /**
     * 项目下项目经理
     *
     * @param projectId 项目ID
     * @return 项目精力列表
     */
    public BocloudResult listManagers(Long projectId) {
        return serviceFactory
                .build(BoCloudService.SMS, "/v1/projects/" + projectId + "/managers", BoCloudMethod.GET, headers, null)
                .invoke();
    }

    public List<Project> projectsByCondition(Long target, Catalog catalog) {
        JSONObject object = new JSONObject();
        object.put(Common.CONDITION, "projectsByCondition");
        object.put("target", target);
        object.put("catalog", catalog.name());
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/projects", BoCloudMethod.GET, headers,
                MapTools.simpleMap(Common.CONDITION, object.toJSONString()));
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (!StringUtils.isNotEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Project.class);
    }

    public List<Project> listTenantProject(Long tenantId) {
        JSONObject object = new JSONObject();
        object.put(Common.CONDITION, "listTenantProject");
        if (null != tenantId && 0 != tenantId) {
            object.put("tenantId", tenantId);
        }
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/projects", BoCloudMethod.GET, headers,
                MapTools.simpleMap(Common.CONDITION, object.toJSONString()));
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (!StringUtils.isNotEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Project.class);
    }

    public List<Tenant> listProjectTenants(Long projectId) {
        JSONObject object = new JSONObject();
        object.put(Common.CONDITION, "listProjectTenants");
        if (null != projectId && 0 != projectId) {
            object.put("projectId", projectId);
        }
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/tenants", BoCloudMethod.GET, headers,
                MapTools.simpleMap(Common.CONDITION, object.toJSONString()));
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (StringUtils.isEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Tenant.class);
    }

    public List<Project> listProjectByUserId(Long userId) {
        JSONObject object = new JSONObject();
        object.put(Common.CONDITION, "listUserProjects");
        object.put("userId", userId);
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/projects", BoCloudMethod.GET, headers,
                MapTools.simpleMap(Common.CONDITION, object.toJSONString()));
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (!StringUtils.isNotEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Project.class);
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        this.headers.put(BocloudToken.INTERNAL_TOKEN, currentService.getToken());
        this.headers.put("source", currentService.getService().getName());
    }

    /**
     * 查询租户下项目
     *
     * @param tenantId
     * @return
     */
    public List<Project> listTenantProjects(Long tenantId) {
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/tenants/" + tenantId + "/projects", BoCloudMethod.GET,
                headers, null);
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (StringUtils.isEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Project.class);
    }

    /**
     * 查询用户管理的项目
     *
     * @param userId
     * @return
     */
    public List<Project> listUserManagerProjects(Long userId) {
        JSONObject object = new JSONObject();
        object.put(Common.CONDITION, "listUserManagerProjects");
        object.put("userId", userId);
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/projects/condition", BoCloudMethod.GET, headers,
                MapTools.simpleMap(Common.CONDITION, object.toJSONString()));
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (StringUtils.isEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Project.class);
    }

    /**
     * 查询用户有可见权限的项目
     *
     * @param userId
     * @return
     */
    public List<Project> listUserVisibleProjects(Long userId) {
        JSONObject object = new JSONObject();
        object.put(Common.CONDITION, "listUserVisibleProjects");
        object.put("userId", userId);
        RemoteService remote = serviceFactory.build(SERVICE, "/v1/projects/condition", BoCloudMethod.GET, headers,
                MapTools.simpleMap(Common.CONDITION, object.toJSONString()));
        BocloudResult result = remote.invoke();
        if (result.isFailed()) {
            return Lists.newArrayList();
        }
        String data = JSONObject.toJSONString(result.getData());
        if (StringUtils.isEmpty(data)) {
            return Lists.newArrayList();
        }
        return JSONArray.parseArray(data, Project.class);
    }

}
